from rebotics_sdk.providers.utils import is_valid_url
from .base import ReboticsBaseProvider, remote_service, ProviderHTTPClientException


class AdminProvider(ReboticsBaseProvider):
    @remote_service('/version/')
    def version(self):
        return self.session.get()

    @remote_service('/admin/', json=False)
    def admin_ping(self, **kwargs):
        return self.session.get()

    def get_retailer_tf_models(self, codename=None):
        if codename is not None and not self.is_retailer_identifier_used():
            return self.get_retailer_tf_models_by_codename(codename)
        elif self.is_retailer_identifier_used():
            return self.get_retailer_tf_models_by_retailer_authentication()
        else:
            raise ProviderHTTPClientException('You did not use any of the authentication methods to get configurations')

    @remote_service('/nn_models/tf/models/')
    def get_retailer_tf_models_by_retailer_authentication(self, **kwargs):
        return self.session.get()

    @remote_service('/nn_models/tf/models/{codename}/')
    def get_retailer_tf_models_by_codename(self, codename, **kwargs):
        return self.session.get(codename=codename)

    @remote_service('/retailers/host/')
    def get_retailer(self, retailer_codename, **kwargs):
        response = self.session.post(data={
            'company': retailer_codename
        })
        return response

    def get_retailer_host(self, retailer_codename):
        return self.get_retailer(retailer_codename)['host']

    @remote_service('/retailers/')
    def get_retailer_list(self, **kwargs):
        return self.session.get()

    @remote_service('/retailers/host/{codename}/')
    def update_host(self, codename, host, **kwargs):
        if not is_valid_url(host):
            raise ProviderHTTPClientException('%s is not a valid url' % host, host=host)
        return self.session.patch(codename=codename, data={
            'host': host
        })

    def set_retailer_identifier(self, retailer_id, retailer_secret_key):
        self.headers['x-retailer-id'] = retailer_id
        self.headers['x-retailer-secret-key'] = retailer_secret_key

    @remote_service('/api/token-auth/')
    def token_auth(self, username, password, **kwargs):
        json_data = self.session.post(data={
            'username': username,
            'password': password
        })
        self.set_token(json_data['token'])
        return json_data

    def get_configurations(self, codename=None):
        """
        Should call this with codename and provided token authentication,
        or
        with provided retailer identifier
        :param codename:
        :return:
        """
        if codename is not None and not self.is_retailer_identifier_used():
            return self.get_configurations_by_codename(codename)
        elif self.is_retailer_identifier_used():
            return self.get_configurations_by_retailer_authentication()
        else:
            raise ProviderHTTPClientException('You did not use any of the authentication methods to get configurations')

    @remote_service('/retailers/host/{codename}/configurations/')
    def get_configurations_by_codename(self, codename, **kwargs):
        return self.session.get(codename=codename)

    @remote_service('/retailers/configurations/')
    def get_configurations_by_retailer_authentication(self, **kwargs):
        return self.session.get()

    # deprecated
    @remote_service('/api/classification_data/import/')
    def create_classification_database_import(self, retailer, model, extension='zip', **kwargs):
        allowed_extensions = ['rcdb', 'zip']
        if extension not in allowed_extensions:
            raise ValueError("Extension should be on of {} not {}".format(
                allowed_extensions, extension
            ))
        return self.session.post(json={
            'retailer': retailer,
            'model': model,
            'extension': extension
        })

    # deprecated
    @remote_service('/api/classification_data/task/export_feature_database/{id}/complete/')
    def notify_classification_database_import_done(self, id, **kwargs):
        return self.session.post(
            id=id,
        )

    # deprecated
    @remote_service('/api/classification_data/task/export_feature_database/{id}/complete/')
    def notify_classification_database_import_failed(self, id, error_message='', **kwargs):
        return self.session.post(
            id=id,
            json={
                'error': error_message
            }
        )

    @remote_service('/api/classification_data/rcdb/')
    def rcdb_create(self, data, **kwargs):
        return self.session.post(json=data)

    @remote_service('/api/classification_data/rcdb/{id}/')
    def rcdb_update(self, id_, data, **kwargs):
        return self.session.patch(id=id_, json=data)

    @remote_service('/api/classification_data/rcdb/{id}/')
    def rcdb_get(self, id_, **kwargs):
        return self.session.get(id=id_)

    @property
    def rcdb(self):
        """

        :return: ReboticsClassificationDatabase utility tools. Higher level of provider
        :rtype: ReboticsClassificationDatabase
        """
        from ..advanced.flows import ReboticsClassificationDatabase
        return ReboticsClassificationDatabase(provider=self)

    @remote_service('/core/core-callback/{id}/')
    def get_core_callback_data(self, pk):
        return self.session.get(id=pk)



from django.contrib.admin.models import LogEntry, CHANGE, ADDITION
from django.contrib.contenttypes.models import ContentType


def log_action(user, model_object, message, create=True):
    LogEntry.objects.log_action(
        user_id=user.id,
        content_type_id=ContentType.objects.get_for_model(model_object).pk,
        object_id=model_object.id,
        object_repr=str(model_object),
        action_flag=ADDITION if create else CHANGE,
        change_message=message,
    )



from rest_framework import serializers


class WebhookDataSerializer(serializers.Serializer):
    status = serializers.CharField()
    component = serializers.CharField()
    body = serializers.JSONField()



import csv
import json
import logging
import os
import webbrowser
from copy import copy
from datetime import datetime
from multiprocessing import Pool
from pathlib import Path

import click
import pandas as pd
from tqdm import tqdm

from rebotics_sdk import utils
from rebotics_sdk.advanced.reverse_planogram import make_planogram_df
from .common import shell, roles, configure
from .renderers import format_processing_action_output, format_full_table
from .tasks import task_create_processing_action_for_image, upload_preview_task, \
    task_download_processing_action, task_recalculate_processing, task_cancel_processing, task_requeue_processing
from .utils import ReboticsCLIContext, pass_rebotics_context, app_dir, process_role, read_saved_role, task_runner, \
    download_file_from_dict, guess_input_type
from ..advanced import remote_loaders
from ..providers import RetailerProvider, ProviderHTTPClientException
from ..utils import mkdir_p, get_filename_from_url, download_file

logger = logging.getLogger(__name__)


@click.group()
@click.option('-f', '--format', default='table', type=click.Choice(['table', 'id', 'json']))
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode')
@click.option('-c', '--config', type=click.Path(), default='retailers.json', help="Specify what config.json to use")
@click.option('-r', '--role', default=lambda: read_saved_role('retailer'), help="Key to specify what retailer to use")
@click.version_option()
@click.pass_context
def api(ctx, format, verbose, config, role):
    """
    Retailer CLI tool to communicate with retailer API
    """
    process_role(ctx, role, 'retailer')
    ctx.obj = ReboticsCLIContext(
        role,
        format,
        verbose,
        os.path.join(app_dir, config),
        provider_class=RetailerProvider
    )


@api.command()
@pass_rebotics_context
def version(ctx):
    """Show retailer backend version"""
    ctx.format_result(ctx.provider.version(), 100)


@api.command()
@click.option('-t', '--input_type')
@click.option('-s', '--store', type=click.INT)
@click.argument('files', nargs=-1, required=True, type=click.File('rb'))
@pass_rebotics_context
def upload_files(ctx, input_type, store, files):
    """
    Upload processing files to the retailer backend, that can be used as processing action inputs
    """
    file_ids = []
    for f_ in files:
        response = ctx.provider.processing_upload(
            store, f_, input_type
        )
        file_ids.append(response['id'])

        if ctx.verbose:
            click.echo(response)  # redirecting this output to stderr
    click.echo(' '.join(map(str, file_ids)))


REQUEUE_TYPES = {
    "facenet_kf": 'requeue_for_facenet_keyframes_key',
    "pt_multiclass": 'REQUEUE_PRICE_TAGS_DETECTION_MULTICLASS',
    "pt_heatmap": 'REQUEUE_PRICE_TAGS_DETECTION_MULTICLASS_HEATMAP',
    "pt_voting": 'REQUEUE_PRICE_TAGS_DETECTION_MULTICLASS_VOTING',
}


@api.command()
@click.argument('processing_ids', required=True, nargs=-1, type=click.INT)
@click.option('-t', '--requeue-type', type=click.Choice(choices=REQUEUE_TYPES.keys()), required=False, default=None)
@click.option('-c', '--concurrency', type=int, default=4)
@pass_rebotics_context
def requeue(ctx, processing_ids, requeue_type, concurrency):
    """Requeue processing actions by given IDs"""
    return task_runner(ctx, task_requeue_processing, processing_ids, concurrency, requeue_type=requeue_type)


@api.command()
@click.argument('processing_ids', required=True, nargs=-1, type=click.INT)
@click.option('-c', '--concurrency', type=int, default=4)
@pass_rebotics_context
def cancel(ctx, processing_ids, concurrency):
    """Cancel processing of the actions by given IDs"""
    return task_runner(ctx, task_cancel_processing, processing_ids, concurrency)


@api.command()
@click.argument('processing_ids', required=True, nargs=-1, type=click.INT)
@click.option('-b', '--batch-size', type=int, default=3,
              help="The lower the number the slower and more reliable the upload will be")
@pass_rebotics_context
def export_to_dataset(ctx, processing_ids, batch_size):
    if ctx.verbose:
        click.echo("Exporting to dataset with batch size {} scans: {}".format(batch_size, processing_ids), err=True)

    try:
        if ctx.verbose:
            click.echo("Sending request...")
        response = ctx.provider.export_to_dataset(processing_ids, batch_size)
        click.echo('Queued to export to dataset: {}'.format(response))
    except ProviderHTTPClientException as exc:
        if ctx.verbose:
            logger.exception(exc)
        raise click.ClickException(str(exc.response.json()))


@api.command()
@click.argument('processing_ids', required=True, nargs=-1, type=click.INT)
@click.option('-c', '--concurrency', type=int, default=4)
@pass_rebotics_context
def recalculate(ctx, processing_ids, concurrency):
    """Recalculate processing of the actions by given IDs"""
    return task_runner(ctx, task_recalculate_processing, processing_ids, concurrency)


@api.command()
@click.option('-t', '--input_type')
@click.option('-s', '--store', type=click.INT)
@click.option('-p', '--store-planogram', type=click.INT)
@click.option('--aisle')
@click.option('--section')
@click.option('-l', '--lens-used', is_flag=True, default=False)
@click.argument('files', nargs=-1, required=True, type=click.INT)
@pass_rebotics_context
def create_processing_action(ctx, input_type, store, store_planogram, aisle, section, lens_used, files):
    """Create processing action for store defining files by IDs"""
    response = ctx.provider.create_processing_action(
        store, files, input_type,
        store_planogram=store_planogram,
        aisle=aisle,
        section=section,
        lens_used=lens_used
    )
    if ctx.verbose:
        click.echo(json.dumps(response, indent=2))
    click.echo(response['id'])


@api.command()
@click.argument('actions', nargs=-1, required=True, type=click.INT)
@click.option('-t', '--target', type=click.Path(), default='.')
@click.option('-c', '--concurrency', type=int, default=4)
@pass_rebotics_context
def download_processing_action(ctx, actions, target, concurrency):
    """Download processing actions by given IDs"""
    files_to_download = []

    pool = Pool(concurrency)
    actions_data = pool.starmap(task_download_processing_action, [(ctx, action) for action in actions])
    if ctx.verbose:
        click.echo('GET API for processing actions completed')

    for data in actions_data:
        action_id = data['id']
        processing_action_folder = os.path.join(target, 'ProcessingAction#%d' % action_id)

        mkdir_p(processing_action_folder)
        results = os.path.join(processing_action_folder, 'results')
        inputs = os.path.join(processing_action_folder, 'inputs')

        mkdir_p(results)
        mkdir_p(inputs)

        for key in ['merged_image_jpeg', 'merged_image', ]:
            files_to_download.append({
                'url': data[key],
                'filepath': os.path.join(results, get_filename_from_url(data[key])),
                'ctx': ctx,
            })

        for input_object in data.get('inputs', []):
            files_to_download.append({
                'filepath': os.path.join(inputs, get_filename_from_url(input_object['file'])),
                'url': input_object['file'],
                'ctx': ctx
            })

        with open(os.path.join(processing_action_folder, 'processing_action_%d.json' % action_id), 'w') as fout:
            json.dump(data, fout, indent=4)

        if ctx.verbose:
            click.echo('Downloading files for %s' % (action_id,))

    pool.map(download_file_from_dict, files_to_download)

    if ctx.verbose:
        click.echo('Processing download success')


@api.command()
@click.argument('actions', nargs=-1, required=True, type=click.INT)
@click.option('-t', '--target', type=click.Path(), default='.')
@click.option('-c', '--concurrency', type=int, default=4)
@click.option('--template', default="{action_id}_{i}.{ext}")
@pass_rebotics_context
def download_processing_inputs(ctx, actions, target, concurrency, template):
    """Download processing inputs from processing actions"""
    files_to_download = []

    pool = Pool(concurrency)
    actions_data = pool.starmap(task_download_processing_action, [(ctx, action) for action in actions])

    if ctx.verbose:
        click.echo('GET API for processing actions completed')

    for data in actions_data:
        for i, input_object in enumerate(data.get('inputs', [])):
            filename = get_filename_from_url(input_object['file'])
            ext = filename.split('.')[-1]
            action_id = data['id']
            files_to_download.append({
                'filepath': os.path.join(target, template.format(action_id=action_id, i=i, ext=ext)),
                'url': input_object['file'],
                'ctx': ctx
            })

    if ctx.verbose:
        click.echo("Task registraion completed")

    pool.map(download_file_from_dict, files_to_download)

    if ctx.verbose:
        click.echo('Processing inputs download success')


@api.command()
@click.option('-d', '--delete', is_flag=True)
@click.option('-c', '--concurrency', type=int, default=4)
@click.argument('target', type=click.Path(exists=True), default=os.getcwd())
@pass_rebotics_context
def upload_previews_from_folder(ctx, delete, concurrency, target):
    """
    Upload previews from file system to the server in parallel.
    It has increased retries and timeout.

    You need to have the file structure like this

\b
target_folder/
└── 6925303739454
    ├── preview_1.png
    ├── preview_2.png
    └── preview_3.png
    """
    provider = ctx.provider
    if provider is None:
        raise click.ClickException('You have supplied role that is not correct!')

    ctx.provider.retries = 5
    ctx.provider.timeout = 300
    verbose = ctx.verbose

    tasks = []
    for label in os.listdir(target):
        upc_folder = os.path.join(target, label)
        if not os.path.isdir(upc_folder):
            continue

        if verbose:
            click.echo('Reading folder: %s' % upc_folder)
        task = {
            'ctx': ctx,
            'images_path': [],
            'delete': delete,
            'upc': label
        }

        if label.isdigit():
            if ctx.verbose:
                click.echo('Registering {} folder'.format(upc_folder))
            for filename in os.listdir(upc_folder):
                image_path = os.path.join(upc_folder, filename)

                if os.path.isfile(image_path):
                    task['images_path'].append(image_path)

        if task['images_path']:
            tasks.append(task)

    if verbose:
        click.echo('Number of tasks: {}'.format(len(tasks)))

    p = Pool(concurrency)
    p.map(upload_preview_task, tasks)
    click.echo('Finished')


PROCESSING_STATUSES = {
    'created': 'action created',
    'done': 'done',
    'error': 'error',
    'interrupted': 'interrupted',
    'progress': "in progress",
}


@api.command()
@click.option('-s', '--store', type=click.INT, help='Store ID')
@click.option('--status', type=click.Choice(PROCESSING_STATUSES.keys()))
@click.option('-p', '--page', type=click.INT, default=1)
@click.option('-r', '--page-size', type=click.INT, default=10)
@pass_rebotics_context
def processing_actions(ctx, store, status, page, page_size):
    """Fetches processing actions and renders them in terminal"""
    if ctx.verbose:
        click.echo('Getting list of processing actions')
    try:
        data = ctx.provider.processing_action_list(
            store,
            PROCESSING_STATUSES.get(status),
            page=page,
            page_size=page_size,
        )
        click.echo('Total results: %d' % len(data))
        if ctx.format != 'id':
            format_processing_action_output(data, 'id')
        format_processing_action_output(data, ctx.format)
    except ProviderHTTPClientException as exc:
        if ctx.verbose:
            logger.exception(exc)
        raise click.ClickException('Failed to get list of processing actions')


@api.command()
@click.option('-t', '--token')
@click.argument('url')
@pass_rebotics_context
def set_webhook(ctx, token, url):
    """Setting webhook url for current user"""
    data = ctx.provider.set_webhook(url, token)
    click.echo('Webhook ID on server is : %d' % data['id'])


@api.command()
@click.option('-t', '--title', help='Planogram title', required=True)
@click.option('-d', '--description', help='Planogram description', default='')
@click.argument('planogram_file', type=click.File(mode='rb'))
@pass_rebotics_context
def import_planogram(ctx, planogram_file, title, description):
    """Upload planogram file to retailer instance in a very specific format"""
    try:
        ctx.provider.import_planogram(planogram_file, title, description)
    except (AssertionError, ProviderHTTPClientException) as exc:
        click.echo(exc, err=True)


@api.command()
@click.argument('planogram_assign_file', type=click.File(mode='r'))
@click.option('-d', '--deactivate', is_flag=True, help='Deactivate old planogram')
@pass_rebotics_context
def assign_planogram_through_file(ctx, planogram_assign_file, deactivate):
    """ Assign Planogram through the file """
    try:
        ctx.provider.assign_planogram(planogram_assign_file, deactivate)
    except (AssertionError, ProviderHTTPClientException) as exc:
        click.echo(exc, err=True)


@api.command()
@click.argument('braincorp_csv', type=click.File(mode='r'))
@click.option('-s', '--store', type=click.INT, help='Store ID')
@click.option('-t', '--target', default='after_processing_upload.csv')
@click.option('-c', '--concurrency', type=int, default=4)
@pass_rebotics_context
def upload_brain_corp_images(ctx, braincorp_csv, store, target, concurrency):
    df = pd.read_csv(braincorp_csv, header=None, names=[
        'date', 'time', 'image', 'x', 'y', 'yaw', 'type', 'count', 'note',
    ])
    pool = Pool(concurrency)

    processing_actions_ids = pool.starmap(task_create_processing_action_for_image, [
        (ctx, store, image_path) for image_path in df['image']
    ])
    df['action_id'] = processing_actions_ids
    click.echo('Processing upload completed. Please check completion using: ')
    click.echo('retailer processing-actions -s {}'.format(store))
    click.echo(df.head())
    df.to_csv(target)
    click.echo('File is written to {}'.format(target))


@api.command()
@click.argument('braincorp_csv', type=click.File(mode='r'))
@click.option('-t', '--target', default='after_processing_completed.csv')
@click.option('-c', '--concurrency', type=int, default=4)
@pass_rebotics_context
def process_brain_corp_images(ctx, braincorp_csv, target, concurrency):
    df = pd.read_csv(braincorp_csv)
    pool = Pool(concurrency)

    processing_actions = pool.starmap(task_download_processing_action, [
        (ctx, action_id) for action_id in df['action_id']
    ])

    def get_product_plu(action):
        if not action:
            return None
        item_upcs = map(lambda x: x['upc'], action['items'])
        return ','.join(set(item_upcs))

    identified_products = list(map(get_product_plu, processing_actions))

    df['identified_products'] = identified_products
    click.echo(df.head())
    df.to_csv(target)
    click.echo('File is written to {}'.format(target))


@api.command()
@click.argument('store_id', type=int)
@pass_rebotics_context
def store_aisles(ctx, store_id):
    """
    This API endpoint returns a list of the aisles and sections for store, accessed by id.
    Allows only GET on detail-route
    Example: get: /api/v4/store/store_planograms/<store_id>/
    """
    aisles = ctx.provider.get_store_aisles(store_id)
    # flatter results
    results = []
    for aisle in aisles:
        sections = aisle.pop('sections')
        for section in sections:
            aisle_section = copy(aisle)
            aisle_section['section'] = section
            results.append(aisle_section)
        else:
            results.append(aisle)

    format_full_table(results)


@api.command()
@pass_rebotics_context
def store_list(ctx):
    """ Return all stores related to the authenticated user"""
    results = ctx.provider.get_stores()
    format_full_table(results)


@api.command()
@click.argument('username')
@pass_rebotics_context
def user_subscriptions(ctx, username):
    """ Returns all subscriptions of the user"""
    ctx.format_result(ctx.provider.user_subscriptions(username))


@api.command()
@click.option('-s', '--store', help='Store ID')
@click.option('-a', '--aisle', help='Aisle')
@click.option('-S', '--section', help='Section')
@click.argument('username')
@pass_rebotics_context
def user_subscribe(ctx, username, store, aisle, section):
    """
    Create a subscription of the user to specific aisle and section scan updates in store alongside with AEON features
    """
    ctx.format_result(ctx.provider.user_subscriptions_create(
        username, store, aisle, section
    ))


def task_upload_and_notify_request(task_def):
    with open(task_def['filepath'], 'rb') as file_io:
        response = remote_loaders.upload(task_def['destination'], file_io)
    status_code = response.status_code

    if 200 <= status_code < 300:
        ctx = task_def['ctx']
        ctx.provider.notify_processing_upload_finished(task_def['id'])
        return task_def['id']
    return None


@api.command()
@click.argument('files', nargs=-1, required=True, type=click.Path(exists=True, dir_okay=False))
@click.option('-c', '--concurrency', type=int, default=4)
@click.option('-t', '--input_type')
@pass_rebotics_context
def upload_processing_files(ctx, files, concurrency, input_type):

    task_defs = []

    for file in tqdm(files, 'Getting S3 pre-signed post', leave=False):
        filepath = Path(file)
        if input_type is None:
            input_type = guess_input_type(filepath.suffix)

        req = ctx.provider.processing_upload_request(filepath.name, input_type)
        task_defs.append({
            'destination': req['destination'],
            'id': req['id'],
            'filepath': file,
            'ctx': ctx,
        })

    pool = Pool(concurrency if concurrency >= len(task_defs) else len(task_defs))
    uploaded = []

    for id_ in tqdm(
        pool.imap_unordered(task_upload_and_notify_request, task_defs),
        'Uploading files', total=len(task_defs), leave=False
    ):
        if id_ is not None:
            uploaded.append(id_)

    click.echo("Uploaded: {}".format(uploaded))
    click.echo("Failed: {}".format(list(
        set(t['id'] for t in task_defs) - set(uploaded)
    )))


@api.group(invoke_without_command=True)
@click.argument('processing_id', required=True, type=click.STRING)
@pass_rebotics_context
@click.pass_context
def processing_action(click_context, ctx, processing_id):
    """Returns processing action by processing_id."""
    try:
        processing_ids = utils.parse_id_range_string(processing_id)
    except ValueError as exc:
        raise click.ClickException(exc)

    setattr(ctx, 'processing_id', processing_ids[0])
    setattr(ctx, 'processing_id_string', processing_id)
    setattr(ctx, 'processing_id_list', processing_ids)

    if click_context.invoked_subcommand is None:
        result = ctx.provider.processing_action_detail(processing_ids[0])
        format_processing_action_output([result, ], ctx.format)


@processing_action.command(name='realogram')
@pass_rebotics_context
def processing_action_realogram(ctx):
    """Returns processing realogram by processing_id."""
    try:
        result = ctx.provider.processing_action_realogram_detail(ctx.processing_id)
        ctx.format_result(result, keys_to_skip=['banner_id'])
    except ProviderHTTPClientException as exc:
        if ctx.verbose:
            logger.exception(exc)
        click.echo('Failed relogram: %s' % exc)


@processing_action.command(name='download')
@click.option('-t', '--target', type=click.Path(), default='.')
@click.option('-c', '--concurrency', type=int, default=4)
@pass_rebotics_context
def processing_action_download(ctx, target, concurrency):
    """Download the processing and returns the result of this processing"""
    files_to_download = []

    data = task_download_processing_action(ctx, ctx.processing_id)
    action_id = data['id']
    processing_action_folder = os.path.join(target, 'ProcessingAction#%d' % action_id)

    mkdir_p(processing_action_folder)
    results = os.path.join(processing_action_folder, 'results')
    inputs = os.path.join(processing_action_folder, 'inputs')

    mkdir_p(results)
    mkdir_p(inputs)

    for key in ['merged_image_jpeg', 'merged_image', ]:
        files_to_download.append({
            'url': data[key],
            'filepath': os.path.join(results, get_filename_from_url(data[key])),
            'ctx': ctx,
        })

    for input_object in data.get('inputs', []):
        files_to_download.append({
            'filepath': os.path.join(inputs, get_filename_from_url(input_object['file'])),
            'url': input_object['file'],
            'ctx': ctx
        })

    with open(os.path.join(processing_action_folder, 'processing_action_%d.json' % action_id), 'w') as fout:
        json.dump(data, fout, indent=4)

    pool = Pool(concurrency)
    pool.map(download_file_from_dict, files_to_download)

    if ctx.verbose:
        click.echo('Processing download success')


@processing_action.command(name='requeue')
@click.option('-t', '--requeue-type', type=click.Choice(choices=REQUEUE_TYPES.keys()), required=False, default=None)
@pass_rebotics_context
def processing_action_requeue(ctx, requeue_type):
    """ Requeue the processing action by processing_id """
    try:
        result = ctx.provider.requeue(ctx.processing_id)
        format_processing_action_output([result, ], ctx.format)
    except ProviderHTTPClientException as exc:
        click.echo('Requeue is not allowed %s' % exc)


@processing_action.command(name='cancel')
@pass_rebotics_context
def processing_action_cancel(ctx):
    """Cancel processing calculation by processing_id"""
    try:
        result = ctx.provider.cancel(ctx.processing_id)
        format_processing_action_output([result, ], ctx.format)
    except ProviderHTTPClientException as exc:
        click.echo('Cancel is not allowed: %s' % exc)


@processing_action.command(name='recalculate')
@pass_rebotics_context
def processing_action_recalculate(ctx):
    """Recalculate processing action by processing_id"""
    try:
        result = ctx.provider.recalculate(ctx.processing_id)
        format_processing_action_output([result, ], ctx.format)
    except ProviderHTTPClientException as exc:
        click.echo('Recalculate is not allowed: %s' % exc)


@processing_action.command(name='view')
@pass_rebotics_context
def processing_action_view_in_admin(ctx):
    """View processing action in admin"""
    url = ctx.provider.build_url('/admin/processing/processingaction/{}/change/'.format(ctx.processing_id))

    if ctx.verbose:
        click.echo('Opening processing action in browser: %s' % url)
    webbrowser.open(url)


@processing_action.command(name='delete')
@pass_rebotics_context
def processing_action_delete(ctx):
    """Delete existing processing action by processing_id"""
    try:
        result = ctx.provider.processing_action_delete(ctx.processing_id)
        click.echo('Successfully deactivated processing action %s', result)
    except ProviderHTTPClientException as exc:
        if ctx.verbose:
            logger.exception(exc)
        click.echo('Failed to deactivate: %s' % exc)


@processing_action.command(name='copy')
@click.option('-s', '--store', type=click.INT)
@pass_rebotics_context
def processing_action_copy(ctx, store):
    """Copy of the existing processing action by processing_id"""

    action = ctx.provider.processing_action_detail(ctx.processing_id)

    store_id = action['store_id']
    if store:
        store_id = store

    inputs_id = [i['id'] for i in action['inputs']]

    try:

        result = ctx.provider.create_processing_action(
            store_id,
            files=inputs_id,
            input_type=action['input_type'],
        )
        click.echo('Successfully created processing action')
        format_processing_action_output([result, ], ctx.format)
    except ProviderHTTPClientException as exc:
        if ctx.verbose:
            logger.exception(exc)
        click.echo('Failed to create processing action: %s' % exc)


@processing_action.command(name='notify-oos')
@click.option('-f', '--force', is_flag=True, default=False)
@pass_rebotics_context
def processing_action_notify_oss_report_ready(ctx, force):
    try:
        result = ctx.provider.send_oos_notification(ctx.processing_id, force)
        ctx.format_result(result)
    except ProviderHTTPClientException as exc:
        if ctx.verbose:
            logger.exception(exc)
        click.echo('Failed to notify OOS report ready. %s' % exc)


@processing_action.command(name="report-image")
@click.option('-l', '--label', type=click.Choice([
    "unique_id",
    "product_unique_id",
    "user_selected_unique_id",
    "original_unique_id"
]), default='unique_id')
@click.option('-t', '--target', type=click.Path(), default='.')
@pass_rebotics_context
def processing_action_report_image(ctx, label, target):
    image_url = ctx.provider.build_url("/api/v4/processing/actions/{id}/report_image/".format(id=ctx.processing_id))
    if ctx.verbose:
        click.echo("URL: {}".format(image_url))

    p = '{}_{}.jpeg'.format(ctx.processing_id, label)
    if os.path.isdir(target):
        p = os.path.join(target, p)

    filepath = download_file(image_url, p, label=label)
    click.echo('Saved to {}'.format(filepath))


@processing_action.command(name='realogram-report')
@click.option('-c', '--concurrency', default=4, help='Define how many parallel workers will be used to access API')
@click.option('-d', '--drop-duplicates', is_flag=True, default=False, help="Doesn't work for now")
@click.option('-t', '--target', type=click.Path(file_okay=False), default='.', help='Define specific folder where'
                                                                                    ' the report files will be saved')
@pass_rebotics_context
def processing_action_realogram_report(ctx, concurrency, drop_duplicates, target):
    """
    Function to download processing actions and extract preview from them.
    It will generate three files: scans.csv, realogram_report.csv, and oos_report.csv from given range or
    comma-separated processing action ids

    Typical usage is:

        retailer processing-action 1000-1005,1010-1015,1017 realogram-report


    Be careful with running the report two times, as it will override the data in folder
     if you have not specified the target folder
    """
    pool = Pool(concurrency)
    processing_action_ids = ctx.processing_id_list
    click.echo('downloading processing actions: {}'.format(processing_action_ids))
    pacs = pool.starmap(task_download_processing_action, [(ctx, id_) for id_ in processing_action_ids])
    processing_actions = [p for p in pacs if p is not None]

    click.echo('Download complete! Generating files...')

    df_scans = pd.DataFrame(dtype=object)
    df_oos = pd.DataFrame(dtype=object)
    df_realogram = pd.DataFrame(dtype=object)

    for processing_action in processing_actions:
        df_bay_reports = pd.DataFrame(processing_action['shelf_bays'], columns=[
            'id',
            'store_planogram_id',
            'aisle',
            'section'
        ], dtype=object)
        realogram = processing_action['items']

        df_realogram_pac = pd.DataFrame(realogram, columns=[
            'unique_id',
            'on_shelf_position', 'display_shelf', 'shelf',
            'position',
            'x_min', 'y_min', 'x_max', 'y_max',
            'shelf_bay_id',
        ], dtype=str)

        df_realogram_pac['processing_action'] = processing_action['id']
        df_realogram_pac['store_id'] = processing_action['store_id']
        df_realogram_pac['store_name'] = processing_action['store_name']
        df_realogram_pac['store_number'] = processing_action['store']['custom_id']
        df_realogram_pac['created'] = processing_action['created']

        df_realogram_pac = df_realogram_pac.merge(
            df_bay_reports, left_on='shelf_bay_id', right_on='id'
        )
        df_realogram = df_realogram.append(df_realogram_pac)

        actions = processing_action.get('report_actions', [])
        oos = [a for a in actions if a['action'] == 'ACTION_ADD']
        df_oos_processing_action = pd.DataFrame(oos, columns=[
            'plu', 'from', 'to', 'from_aisle', 'to_aisle',
            'shelf_bay_id',
        ], dtype=object)

        df_oos_processing_action['processing_action'] = processing_action['id']
        df_oos_processing_action['store_id'] = processing_action['store_id']
        df_oos_processing_action['store_name'] = processing_action['store_name']
        df_oos_processing_action['store_number'] = processing_action['store'].get('custom_id')
        df_oos_processing_action['created'] = processing_action['created']
        df_oos_processing_action = df_oos_processing_action.merge(
            df_bay_reports, left_on='shelf_bay_id', right_on='id'
        )
        df_oos = df_oos.append(df_oos_processing_action)

        df_pac = pd.DataFrame([processing_action, ], columns=[
            'id', 'store_name', 'store_id', 'created', 'used_in_report',
            'category_name', 'category_number',
            'store_planogram_id', 'aisle', 'section',
            'download_report_url',
        ], index=['id'], dtype=object)
        df_pac['user'] = processing_action['user'].get('username', '')
        try:
            df_pac['compliance_rate'] = processing_action['compliance_rates'].get('compliance', '0')
        except:
            df_pac['compliance_rate'] = 0
        df_pac['store_number'] = processing_action['store'].get('custom_id')
        df_scans = df_scans.append(df_pac)

    # if drop_duplicates:
    #     TODO: here should be user definable date or splittable. Extend the API
    # df_oos = df_oos.drop_duplicates(['store_number', 'plu', 'created'])

    dump_path = os.path.join(target, 'reporting_{}'.format(ctx.processing_id_string))
    mkdir_p(dump_path)

    df_realogram.to_csv(
        os.path.join(dump_path, 'realogram_report.csv'),
        quoting=csv.QUOTE_ALL,
        index=False
    )
    df_oos.to_csv(
        os.path.join(dump_path, 'oos_report.csv'),
        quoting=csv.QUOTE_ALL,
        index=False
    )
    df_scans.to_csv(
        os.path.join(dump_path, 'scans.csv'),
        quoting=csv.QUOTE_ALL,
        index=False
    )

    click.echo('Written to {}'.format(dump_path))


@processing_action.command()
@click.option('-t', '--target', type=click.Path(file_okay=False), default='.', help='Define specific folder where'
                                                                                    ' the reverse planogram file will be saved')
@pass_rebotics_context
def reverse_planogram(ctx, target):
    if ctx.verbose:
        click.echo("Downloading scan with id #{}".format(ctx.processing_id))

    scan = ctx.provider.processing_action_detail(ctx.processing_id)
    df = make_planogram_df(scan)

    p = 'reverse_planogram_{}.csv'.format(ctx.processing_id)
    if os.path.isdir(target):
        p = os.path.join(target, p)
    df.to_csv(p, index=False, quoting=csv.QUOTE_ALL)
    click.echo("Saved into {}".format(p))


@processing_action.command('spacial')
@click.option('-t', '--target', type=click.Path(file_okay=True, dir_okay=True), default=None,
              help="Save into file, if none provided, will print to STDOUT")
@pass_rebotics_context
def processing_spacial(ctx, target):
    data = ctx.provider.get_spacial_data(ctx.processing_id)

    if target is None:
        # printing to STDOUT
        click.echo(json.dumps(data, indent=2))
        return

    target = Path(target)
    if target.is_dir():
        target /= "spacial_{}_{}.json".format(ctx.processing_id, datetime.now().strftime("%Y-%m-%d"))
    with open(target, 'w') as fout:
        json.dump(data, fout)
    click.echo("Saved into {}".format(target))


@api.group(name='fv')
def feature_vectors():
    """
    Feature vector flow
    """
    pass


@feature_vectors.command('list')
@pass_rebotics_context
def feature_vectors_list(ctx):
    fve = ctx.provider.feature_vectors_export()
    data = fve.get()
    if data:
        ctx.format_result(data)


@feature_vectors.command('export')
@click.option('-b', '--batch-size', type=click.INT, default=50000)
@click.option('-s', '--source-model', default='')
@click.option('-r', '--result-model', default='previews-backup')
@pass_rebotics_context
def feature_vectors_export(ctx, source_model, result_model, batch_size):
    fve = ctx.provider.feature_vectors_export()
    result = fve.export(source_model=source_model, result_model=result_model, batch_size=batch_size)
    ctx.format_result(result)


api.add_command(shell, 'shell')
api.add_command(roles, 'roles')
api.add_command(configure, 'configure')



import os
import webbrowser
from collections import namedtuple

import click

from rebotics_sdk.cli.common import shell, roles, configure, run
from rebotics_sdk.cli.utils import read_saved_role, process_role, ReboticsCLIContext, app_dir, pass_rebotics_context, \
    GroupWithOptionalArgument
from rebotics_sdk.providers import ProviderHTTPServiceException
from rebotics_sdk.providers.shelf_camera_manager import ShelfCameraManagerProvider


@click.group()
@click.option('-f', '--format', default='table', type=click.Choice(['table', 'id', 'json']), help='Result rendering')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode')
@click.option('-c', '--config', type=click.Path(), default='shelf_camera.json', help="Specify what config.json to use")
@click.option('-r', '--role', default=lambda: read_saved_role('shelf_camera'), help="Key to specify what admin to use")
@click.version_option()
@click.pass_context
def api(ctx, format, verbose, config, role):
    """
    Admin CLI tool to communicate with dataset API
    """
    process_role(ctx, role, 'shelf_camera')
    ctx.obj = ReboticsCLIContext(
        role,
        format,
        verbose,
        os.path.join(app_dir, config),
        provider_class=ShelfCameraManagerProvider
    )


CameraType = namedtuple('CameraType', ['type', 'display', 'short'])
CAMERA_TYPES = [
    CameraType(0, 'Shelf Camera', 'shelf'),
    CameraType(1, 'Robot Camera', 'robot'),
    CameraType(2, 'End cap Camera', 'endcap'),
]


@api.group(cls=GroupWithOptionalArgument)
@click.argument('camera_settings_id', required=False)
@pass_rebotics_context
def camera_settings(ctx, camera_settings_id=None):
    if camera_settings_id:
        setattr(ctx, 'camera_settings_id', camera_settings_id)


@camera_settings.command(name='list')
@pass_rebotics_context
def camera_settings_list(ctx):
    ctx.format_result(ctx.provider.camera_settings_list(), keys_to_skip=['x_min', 'y_min', 'x_max', 'y_max'])


@camera_settings.command(name='detail')
@pass_rebotics_context
def camera_settings_detail(ctx):
    camera_settings_id = getattr(ctx, 'camera_settings_id', None)

    camera_settings = ctx.provider.camera_settings_detail(camera_settings_id)
    ctx.format_result(
        camera_settings,
        keys_to_skip=['x_min', 'y_min', 'x_max', 'y_max']
    )
    click.echo('Region of interest')
    ctx.format_result(camera_settings['region_of_interests'])


@camera_settings.command(name='view')
@pass_rebotics_context
def camera_settings_view(ctx):
    camera_settings_id = getattr(ctx, 'camera_settings_id', None)
    url = ctx.provider.build_url('/admin/fetcher/shelfcamerasettings/{}/change/'.format(camera_settings_id))

    if ctx.verbose:
        click.echo('Opening camera settings in browser: %s' % url)

    webbrowser.open(url)


@camera_settings.command(name='create')
@click.option('-t', '--token', help='Auth token for user to be used to send user')
@click.option('-s', '--store', help='Store ID')
@click.option('-r', '--retailer', help='Retailer codename')
@click.option('--rotation', help='Rotation: 0, 90, 180, 270', default=0)
@click.option('-t', '--camera_type', help='Camera type',
              type=click.Choice([t.short for t in CAMERA_TYPES]))
@click.argument('folder')
@pass_rebotics_context
def create_camera_settings(ctx, token, store, retailer, camera_type, folder, rotation):
    """Create camera settings with binding of stores and users and folder"""
    camera_type_codes = [c for c in CAMERA_TYPES if c.short == camera_type]
    if not camera_type_codes:
        raise click.ClickException('Camera type was not found')
    ct = camera_type_codes[0]
    try:
        if ctx.verbose:
            click.echo('Calling create camera settings')
        result = ctx.provider.add_camera_settings(
            retailer,
            store,
            token,
            ct.type,
            folder=folder,
            rotation=rotation
        )
        ctx.format_result(result, keys_to_skip=['x_min', 'y_min', 'x_max', 'y_max'])
        click.echo('Successfully added camera settings with id %d' % result['id'])
    except ProviderHTTPServiceException as exc:
        raise click.ClickException(str(exc))


@camera_settings.command(name='add-roi')
@click.option('-a', '--aisle')
@click.option('-s', '--section')
@click.option('-p', '--planogram_id')
@click.option('-c', '--coordinates')
@pass_rebotics_context
def camera_settings_add_roi(ctx, aisle, section, planogram_id, coordinates):
    camera_settings_id = getattr(ctx, 'camera_settings_id', None)
    if camera_settings_id is None:
        raise click.ClickException('You need to specify camera_settings_id. Usage: \n'
                                   'camera_manager camera-settings 123 add-roi')

    ctx.format_result(
        ctx.provider.add_camera_roi(
            camera_settings_id,
            aisle=aisle,
            section=section,
            store_planogram_id=planogram_id,
        )
    )


api.add_command(shell, 'shell')
api.add_command(roles, 'roles')
api.add_command(configure, 'configure')
api.add_command(run, 'run')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-06-05 08:16
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0002_auto_20190527_0938'),
    ]

    operations = [
        migrations.AlterField(
            model_name='message',
            name='component',
            field=models.CharField(choices=[(b'processing', 'Processing action'), (b'planogram_import', 'Planogram import'), (b'internal', 'Internal')], max_length=64),
        ),
        migrations.AlterField(
            model_name='message',
            name='delivery_status',
            field=models.CharField(choices=[(b'message_pending', 'Pending'), (b'message_delivered', 'Delivered'), (b'message_rejected', 'Rejected'), (b'message_failed', 'Failed')], max_length=64),
        ),
    ]



import logging

import requests
from celery import shared_task
from rest_framework import status

from .models import Message

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 180


@shared_task(bind=True)
def send_message(self, message_id):
    message = Message.objects.get(id=message_id)
    try:
        if message.receiver.auth_token:
            headers = {
                'Authorization': 'Token %s' % message.receiver.auth_token
            }
        else:
            headers = {}

        response = requests.post(message.receiver.url, data={
            'component': message.component,
            'status': message.status,
            'body': message.body,  # requires JSON string
        }, timeout=DEFAULT_TIMEOUT, headers=headers)

        if response.status_code == status.HTTP_202_ACCEPTED:
            logger.info('Message of component %s is posted to %s', message.component, message.receiver.url)
            message.delivery_status = Message.DELIVERY_STATUS_DELIVERED
        else:
            logger.error('Webhook failed with status %s and message %s', response.status_code, response.content)
            message.delivery_status = Message.DELIVERY_STATUS_REJECTED
            message.exception_message = response.content

    except Exception as exc:
        logger.exception(exc)

        message.delivery_status = Message.DELIVERY_STATUS_FAILED
        message.exception_message = str(exc)
    finally:
        message.save()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-06-07 12:55
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0003_auto_20190605_0816'),
        ('notifications', '0003_auto_20190605_0817'),
    ]

    operations = [
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-27 06:02
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='WebhookRouter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.URLField()),
                ('auth_token', models.CharField(editable=False, max_length=64)),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('modified', models.DateTimeField(auto_now=True, verbose_name='Date Modified')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='webhooks', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Webhook router',
                'verbose_name_plural': 'Webhook routers',
            },
        ),
    ]



import json
import logging
import re

import six

from .base import ReboticsBaseProvider, remote_service

logger = logging.getLogger(__name__)


class DatasetProvider(ReboticsBaseProvider):
    @remote_service('/api/v1/')
    def api_root(self):
        return self.session.get()

    def validate_feature(self, feature_vector):
        if len(feature_vector) != 512:
            raise ValueError('Feature length is inconsistent')

    def clean_features(self, features):
        if isinstance(features, six.string_types):
            return features
        elif isinstance(features, list):
            for feature in features:
                self.validate_feature(feature)
            else:
                return json.dumps(features)

    @remote_service('/api/v1/classifications/classify/', timeout=120)
    def classify(self, facenet, features, threshold=0.6, closeness_range=0.03):
        return self.session.post(data=dict(
            facenet=facenet,
            features=self.clean_features(features),
            threshold=threshold,
            closeness_range=closeness_range
        ))

    @remote_service('/api/v1/classifications/classify_width_dims/', timeout=120)
    def classify_with_dimensions(self, facenet, features, threshold=0.6, closeness_range=0.03):
        return self.session.post(data=dict(
            facenet=facenet,
            feature=self.clean_features(features),
            threshold=threshold,
            closeness_range=closeness_range
        ))

    @remote_service('/api/v1/classifications/classify_width_price_tags_dims/', timeout=120)
    def classify_with_dimensions_and_price_tags(self, facenet, features, price_tags, classification_policy):
        return self.session.post(data=dict(
            facenet=facenet,
            features=self.clean_features(features),
            price_tags=price_tags,
            classification_policy=classification_policy
        ))

    @remote_service('/api/v1/classifications/reference-codes/batch/')
    def save_feature_vector_batch(self, data):
        assert isinstance(data, list), "Should be a list of data, %s type supplied" % type(data)
        return self.session.post(json=data)

    @remote_service('/api/v1/classifications/reference-codes/')
    def save_feature_vector_single(self, data):
        return self.session.post(data=data)

    def inject_retailer_reference(self, data, retailer):
        if 'x-retailer-id' not in self.get_provider_headers().keys():
            if retailer is None:
                raise ValueError('You need to specify retailer '
                                 'if you are not using retailer-in-admin authentication')
            data['retailer'] = retailer.strip()
        return data

    def construct_reference_entry(self, upc, feature, threshold,
                                  retailer, facenet):
        """
        Make dict for threshold
        :param str upc:
        :param str feature:
        :param float threshold:
        :param str retailer:
        :param str facenet:
        :return:
        """
        data = {
            'upc': upc.strip(),
            'feature': feature.strip(),
            'facenet': facenet.strip(),
            'threshold': threshold,
        }

        self.inject_retailer_reference(data, retailer)
        return data

    def save_feature_vector(self, upc, facenet, feature, threshold=0.3, retailer=None, batch_size=1):
        if batch_size > 1:
            raise NotImplementedError(
                'Implement batch upload yourself using provider.save_feature_vector_batch(list[data])'
            )

        return self.save_feature_vector_single(
            data=self.construct_reference_entry(upc, feature, threshold, retailer, facenet)
        )

    @remote_service('/api/v1/classifications/reference-codes/delete/')
    def delete_feature_vector(self, upc, facenet_version, feature, retailer=None):
        data = self.inject_retailer_reference({
            'upc': upc,
            'facenet': facenet_version,
            'feature': feature,
        }, retailer)
        return self.session.post(data=data)

    @remote_service('/api/v1/classifications/reference-codes/{object_id}/')
    def delete_feature_vector_by_id(self, object_id):
        response = self.session.delete(object_id=object_id)
        return response

    @remote_service('/api/v1/token-auth/')
    def token_auth(self, username, password):
        json_data = self.session.post(data=dict(
            username=username,
            password=password
        ))
        self.headers['Authorization'] = 'Token %s' % json_data['token']
        return json_data

    @remote_service('/api/v1/classifications/reference-codes/backup/', raw=True)
    def download_reference_database(self, retailer, facenet):
        response = self.session.post(data=dict(
            retailer=retailer,
            facenet=facenet,
        ), stream=True)
        filename = re.findall("filename=(.+)", response.headers['content-disposition'])[0].strip('"')
        return filename, response.raw

    @remote_service('/api/v1/retailer/settings/features/limit/')
    def set_feature_limit_for_product(self, limit, upc, retailer=None):
        data = dict(
            upc=upc,
            limit=limit
        )
        self.inject_retailer_reference(data, retailer)
        response = self.session.post(data=data)
        return response

    @remote_service('/api/v1/retailer/settings/')
    def set_settings(self, data, retailer=None):
        data = self.inject_retailer_reference(data, retailer)
        response = self.session.post(data=data)
        return response

    @remote_service('/api/v1/check_feature_vector_exists/', timeout=300000)
    def test_feature_vector_existence_bulk(self, data, retailer=None):
        data = self.inject_retailer_reference(data, retailer)
        response = self.session.post(json=data)
        return response

    @remote_service('/api/v1/import_training_data/', timeout=10000)
    def import_training_data(self, data):
        return self.session.post(json=data)

    @remote_service('/api/v1/classifications/reference-codes-read/')
    def get_feature_fectors_by_reatailer_and_model(self, retailer, facenet_model, page=None):
        return self.session.get(
            params={'product__retailer__code': retailer, 'facenet_model': facenet_model, 'page': page})

    @remote_service('/api/v1/classifications/reference-codes/{id}/')
    def update_upc_for_feature_vector(self, fv_id, upc):
        return self.session.patch(id=fv_id, data={'upc': upc})



from rest_framework import serializers

from rebotics_sdk.notifications.models import WebhookRouter


class SetWebhookSerializer(serializers.ModelSerializer):
    class Meta:
        model = WebhookRouter
        fields = [
            'url',
            'auth_token'
        ]
        extra_kwargs = {
            'auth_token': {'required': False},
        }



# coding=utf-8
import csv
import datetime
import json
import os
import pathlib
from functools import partial
from multiprocessing.pool import Pool

import click
import humanize
import tqdm
from prettytable import PrettyTable

from rebotics_sdk.advanced import remote_loaders
from rebotics_sdk.cli.common import shell, configure, roles
from .renderers import format_full_table
from .utils import ReboticsCLIContext, app_dir, pass_rebotics_context, read_saved_role, process_role
from ..advanced.flows import FileUploadError, download_vrcdb_file, ReboticsClassificationDatabase
from ..advanced.packers import ClassificationDatabasePacker, ClassificationDatabaseException, \
    DuplicateFeatureVectorsException, VirtualClassificationDatabasePacker
from ..constants import RCDB
from ..providers import AdminProvider, RetailerProvider, sleep
from ..providers.facenet import FacenetProvider
from ..utils import Timer, uri_validator, download_and_return_id, get_filename_from_url


@click.group()
@click.option('-f', '--format', default='table', type=click.Choice(['table', 'id', 'json']), help='Result rendering')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode')
@click.option('-c', '--config', type=click.Path(), default='admin.json', help="Specify what config.json to use")
@click.option('-r', '--role', default=lambda: read_saved_role('admin'), help="Key to specify what admin to use")
@click.version_option()
@click.pass_context
def api(ctx, format, verbose, config, role):
    """
    Admin CLI tool to communicate with dataset API
    """
    process_role(ctx, role, 'admin')
    ctx.obj = ReboticsCLIContext(
        role,
        format,
        verbose,
        os.path.join(app_dir, config),
        provider_class=AdminProvider
    )


def get_retailer_version_task(retailer_dict):
    retailer_provider = RetailerProvider(host=retailer_dict['host'], retries=1, timeout=5)
    try:
        response = retailer_provider.version()
        version = response['version']
        uptime = humanize.naturaldelta(datetime.timedelta(seconds=int(response['uptime'])))
    except Exception:
        version = 'not working'
        uptime = '---'

    d = [
        retailer_dict['codename'],
        retailer_dict['title'],
        version,
        uptime,
        retailer_dict['host'],
    ]
    return d


@api.command()
@click.argument('retailer')
@pass_rebotics_context
def configurations(ctx, retailer):
    """
        Fetch retailer configurations variables
    """
    try:
        res = ctx.provider.get_configurations(retailer)
        ctx.format_result(res)
    except Exception as exc:
        raise click.ClickException(exc)


@api.command()
@click.option('-n', '--notify', is_flag=True)
@click.option('-d', '--delay', type=click.INT, default=60)
@pass_rebotics_context
def retailer_versions(ctx, notify, delay):
    """Fetch retailer versions and their meta information"""
    click.echo("For better experience use: \n"
               "https://versions.fyn.rocks/ and https://tracker.fyn.rocks/")

    if notify:
        if ctx.verbose:
            click.echo('Using notify option', err=True)

        try:
            from pynotifier import Notification
            Notification(
                title='Subscribed to the notifications',
                description='You will receive notifications for retailer updates',
            ).send()
        except ImportError:
            raise click.ClickException("You can't use notify function")

    provider = ctx.provider
    if ctx.verbose:
        click.echo('Fetching info from rebotics admin.', err=True)
    retailers = provider.get_retailer_list()
    prev_results = []
    results = []
    pool = Pool(len(retailers))

    while True:
        try:
            if ctx.verbose:
                click.echo('Fetching the retailer versions', err=True)
            results = pool.map(get_retailer_version_task, retailers)

            if not notify:
                break

            for prev_result in prev_results:
                retailer_codename = prev_result[0]
                previous_version = prev_result[2]
                for result in results:
                    if result[0] == retailer_codename:
                        current_version = result[2]
                        if previous_version != current_version:
                            notification_message = 'Retailer {} updated from version {} to {}'.format(
                                retailer_codename,
                                previous_version,
                                current_version
                            )
                            click.echo(notification_message)
                            Notification(
                                title=notification_message,
                                description='Current uptime is: {}'.format(result[3]),
                                duration=30,
                                urgency=Notification.URGENCY_CRITICAL,
                            ).send()
            del prev_results
            prev_results = results
            sleep(delay)
        except KeyboardInterrupt:
            break

    table = PrettyTable()
    table.field_names = ['codename', 'title', 'version', 'uptime', 'host']
    for result in results:
        table.add_row(result)
    click.echo(table)


@api.command()
@click.argument('retailer')
@pass_rebotics_context
def models(ctx, retailer):
    """
        Fetch and display retailer NN configurations
    """
    try:
        ctx.format_result(
            ctx.provider.get_retailer_tf_models(codename=retailer)
        )
    except Exception as exc:
        raise click.ClickException(exc)


@api.command()
@click.option('-r', '--retailer', help='Retailer codename')
@click.option('-u', '--facenet-url', help='Facenet service URL')
@click.argument('image_url')
@pass_rebotics_context
def extract_feature_vectors(ctx, retailer, facenet_url, image_url):
    """Fetches latest configuration of neural model for retailer by it's ID and Secret key;
    Sends image to facenet to load model into state."""
    models = ctx.provider.get_retailer_tf_models(codename=retailer)

    facenet_model = models['facenet_model']
    if ctx.verbose:
        click.echo("Facenet model: %s" % facenet_model['codename'])

    facenet_provider = FacenetProvider(facenet_url)
    feature_extractor = partial(
        facenet_provider.extract_from_image_url,
        model_path=facenet_model['data_path'],
        index_path=facenet_model['index_path'],
        meta_path=facenet_model['meta_path'],
    )

    with Timer() as t:
        result = feature_extractor(image_url)
        click.echo(result)

    if ctx.verbose:
        click.echo("Elapsed: %s seconds" % t.elapsed_secs)


@api.command()
@click.option('-r', '--retailer', help='Retailer')
@click.option('-u', '--facenet-url', help='Facenet service URL')
@click.argument('image_url')
@click.argument('bounding_boxes', type=click.File())
@pass_rebotics_context
def extract_feature_vectors_for_boxes(ctx, retailer, facenet_url, image_url, bounding_boxes):
    """Fetches latest configuration of neural model for retailer by it's ID and Secret key;
    Sends keyframe image url and list of bounding boxes to facenet to load model into state."""
    models = ctx.get_retailer_tf_models(codename=retailer)

    facenet_model = models['facenet_model']
    if ctx.verbose:
        click.echo("Facenet model: %s" % facenet_model['codename'])

    boxes = json.load(bounding_boxes)
    assert isinstance(boxes, list), "Need to supply list of bounding boxes"

    facenet_provider = FacenetProvider(facenet_url)
    feature_extractor = partial(
        facenet_provider.extract_from_keyframe,
        model_path=facenet_model['data_path'],
        index_path=facenet_model['index_path'],
        meta_path=facenet_model['meta_path'],
    )

    with Timer() as t:
        result = feature_extractor(keyframe_url=image_url, bboxes=boxes)
        click.echo(result)

    if ctx.verbose:
        click.echo("Elapsed: %s seconds" % t.elapsed_secs)


@api.command()
@click.argument('retailer', type=click.STRING)
@click.argument('url', type=click.STRING)
@pass_rebotics_context
def set_retailer_url(ctx, retailer, url):
    try:
        ctx.provider.update_host(retailer, url)
    except Exception as exc:
        raise click.ClickException(str(exc))
    else:
        click.echo('Set new host for retailer %s' % retailer)


@api.command()
@click.option('-r', '--retailer', help="Retailer codename", prompt=True)
@click.option('-m', '--model', help="Model codename", prompt=True)
@click.option('-i', '--images', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option('-l', '--labels', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-f', '--features', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-c', '--check-duplicates', is_flag=True, help="MD5 checker simple")
@pass_rebotics_context
def pack_and_import_classification_db(ctx, retailer, model, images, labels, features, check_duplicates):
    """Pack classification database to single .rcdb file and import it to the Rebotics Admin

\b
db/
├── custom_folder
│   ├── image_2.png
│   └── image_1.png
├── features.txt
└── labels.txt

    It is a single step command with equivalent of running two commands:

admin pack-classification-db --features features.txt --labels labels.txt --images ./custom_folder/ --target classification.rcdb

admin import_classification_db --retailer delta --model test_code classification.rcdb
    """
    ext = ClassificationDatabasePacker.extension
    with ctx.provider.rcdb.import_flow(retailer, model, ext) as flow:
        if ctx.verbose:
            click.echo("Creating import request: \n"
                       "retailer: {}\n"
                       "model: {}\n"
                       "extension: {}".format(retailer, model, ClassificationDatabasePacker))

        packer = ClassificationDatabasePacker(destination=None, progress_bar=True, check_duplicates=check_duplicates)
        if ctx.verbose:
            click.echo("Packing from provided values: \n"
                       "labels: {labels} \n"
                       "features: {features} \n"
                       "images: {images}".format(labels=labels, features=features, images=images))

        try:
            packed = packer.pack(labels, features, images)
            packed.seek(0)
        except DuplicateFeatureVectorsException as exc:
            handle_duplicates(ctx, exc.duplicates)
            raise click.ClickException(str(exc))
        except ClassificationDatabaseException as exc:
            raise click.ClickException(str(exc))

        if ctx.verbose:
            click.echo("Import request: {}".format(flow.import_request))

        try:
            flow.upload_file(packed, progress_bar=True)
        except FileUploadError as exc:
            click.echo(str(exc))
            raise

        click.echo("Import completed. Reference database id: {}".format(flow.import_request['id']))


@api.command()
@click.option('-r', '--retailer', help="Retailer codename", prompt=True)
@click.option('-m', '--model', help="Model codename", prompt=True)
@click.argument('filepath', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@pass_rebotics_context
def import_classification_db(ctx, retailer, model, filepath):
    """Import rcdb file to Rebotics Admin. Example usage:

    admin import_classification_db --retailer delta --model test_code classification.rcdb

    """
    extension = os.path.split(filepath)[-1].split('.')[-1]
    if ctx.verbose:
        click.echo("Creating import request: \n"
                   "retailer: {}\n"
                   "model: {}\n"
                   "extension: {}".format(retailer, model, extension))

    with ctx.provider.rcdb.import_flow(retailer, model, extension) as flow:
        if ctx.verbose:
            click.echo("Import request: {}".format(flow.import_request))
        with open(filepath, 'rb') as packed:
            flow.upload_file(packed, progress_bar=True)

    click.echo("Import completed. Reference database id: {}".format(flow.import_request['id']))


def handle_duplicates(ctx, duplicates):
    unpacked = [
        [
            {
                'id': entry.index,
                'label': entry.label,
                'filename': entry.filename
            }
            for entry in group
        ] for group in duplicates
    ]
    if ctx.format == 'json':
        click.echo(json.dumps(unpacked, indent=2))
    elif ctx.format == 'table':
        for group in unpacked:
            format_full_table(group)
    elif ctx.format == 'id':
        click.echo("Duplicate id. New line separated for group; space separated inside of the group:")
        for group in duplicates:
            click.echo(" ".join([str(entry.index) for entry in group]))
    else:
        for group in duplicates:
            for entry in group:
                click.echo("{entry.index} {entry.label} {entry.filename}".format(entry=entry))
            click.echo('=' * 100)


@api.command()
@click.option('-i', '--images', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option('-l', '--labels', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-f', '--features', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-c', '--check-duplicates', is_flag=True, help="MD5 checker simple")
@click.argument('target', type=click.Path(exists=False, file_okay=True, dir_okay=True), default='.')
@pass_rebotics_context
def pack_classification_db(ctx, images, labels, features, target, check_duplicates):
    """Pack classification database to single .rcdb file

\b
db/
├── custom_folder
│   ├── image_2.png
│   └── image_1.png
├── features.txt
└── labels.txt

    Example usage:
        admin pack-classification-db --features features.txt --labels labels.txt --images ./custom_folder/ --target classification.rcdb

    """
    target = pathlib.Path(target)
    if target.is_dir():
        target = target / "classification_db{}.{}".format(
            datetime.datetime.now().strftime('%Y-%m-%dZ%H%M'),
            ClassificationDatabasePacker.extension
        )
    packer = ClassificationDatabasePacker(
        destination=target,
        progress_bar=True,
        check_duplicates=check_duplicates
    )
    if ctx.verbose:
        click.echo("Packing from provided values: \n"
                   "labels: {labels} \n"
                   "features: {features} \n"
                   "images: {images}".format(labels=labels, features=features, images=images))
    try:
        packed = packer.pack(labels, features, images)
    except DuplicateFeatureVectorsException as exc:
        handle_duplicates(ctx, exc.duplicates)
        raise click.ClickException(str(exc))
    except ClassificationDatabaseException as exc:
        raise click.ClickException(str(exc))
    click.echo('Written to {}'.format(packed))


@api.group(name='rcdb')
def rcdb_group():
    pass


@rcdb_group.command()
@click.option('-r', '--retailer', help="Retailer codename", prompt=True)
@click.option('-m', '--model', help="Model codename", prompt=True)
@click.option('-c', '--concurrency', type=int, default=4)
@click.argument('target', type=click.Path(exists=False, file_okay=False, dir_okay=True), default='.')
@pass_rebotics_context
def backup_and_download_rcdb(ctx, retailer, model, target, concurrency):
    """
    CLI command to trigger the upload
    """
    rcdb = ReboticsClassificationDatabase(ctx.provider)
    # a flow that will request to do a backup of the retailer + model from the Admin
    res = rcdb.create(retailer, model)
    id_ = res['id']

    # next flow wait for it to be processed calling the function to be done or not every N minutes
    with Timer() as t:
        while True:
            res = rcdb.get(id_)
            if res['status'] == RCDB.EXPORT_DONE:
                file_url = res['backup']
                break
            if res['status'] == RCDB.EXPORT_ERROR:
                raise click.ClickException("Creating and RCDB file failed")
            sleep(5)

    click.echo("Elapsed time: {}".format(t.elapsed_secs))

    # flow that will download the rcdb file
    file_io = remote_loaders.download(file_url, progress_bar=True)

    # next flow that will unpack rcdb into target directory
    target = pathlib.Path(target) / "rcdb_{}_{}_{}".format(retailer, model, id_)

    packer = VirtualClassificationDatabasePacker(source=file_io)
    download_vrcdb_file(packer, target, concurrency=concurrency)


@rcdb_group.command(name='virtual-download')
@click.option('-c', '--concurrency', type=int, default=4)
@click.argument('target', type=click.Path(exists=False, file_okay=False, dir_okay=True), default='.')
@click.argument('backup_ids', type=click.INT, nargs=-1)
@pass_rebotics_context
def download_rcdb(ctx, backup_ids, target, concurrency):
    """
    Download rcdb by id
    :param ctx:
    :param backup_ids:
    :param target:
    :param concurrency:
    :return:
    """
    rcdb = ReboticsClassificationDatabase(ctx.provider)
    target = pathlib.Path(target)

    for backup_id in backup_ids:
        res = rcdb.get(backup_id)

        file_url = res['backup']
        retailer = res['retailer']
        model = res['facenet_model']

        # flow that will download the rcdb file
        file_io = remote_loaders.download(file_url, progress_bar=True)

        # next flow that will unpack rcdb into target directory
        download_folder = target / "rcdb_{}_{}_{}".format(retailer, model, backup_id)

        packer = VirtualClassificationDatabasePacker(source=file_io)
        download_vrcdb_file(packer, download_folder, concurrency=concurrency)


@rcdb_group.command('virtual-unpack')
@click.option('-c', '--concurrency', type=int, default=4)
@click.argument('backup', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument('target', type=click.Path(exists=False, file_okay=False, dir_okay=True), default='.')
@pass_rebotics_context
def unpack_rcdb(backup, target, concurrency):
    # TODO: check by filename and extension
    packer = VirtualClassificationDatabasePacker(source=backup)
    download_vrcdb_file(packer, target, concurrency=concurrency)


@rcdb_group.command(name="previews-fs")
@click.argument("argument", type=click.STRING)
@click.option('-o', "--output", type=click.Path(exists=True, dir_okay=True, file_okay=False), default=".")
@click.option('-c', '--concurrency', default=8, type=click.INT)
@pass_rebotics_context
def rcdb_previews_fs(ctx, argument, output, concurrency):
    """
    Prepare previews file structure by the id, filepath or file url of the JSON formatted backup
    Make sure that extension is a valid JSON and extension in the admin is set like that
    """
    output = pathlib.Path(output).absolute()
    output.mkdir(parents=True, exist_ok=True)

    filepath = output / "definition.json"

    if argument.isidigit():
        backup_id = int(argument)
        rcdb_data = ctx.provider.rcdb.get(backup_id)
        if rcdb_data['extension'] != RCDB.JSON:
            click.echo("extension is not supported")  # but we will try anyway
            # raise click.ClickException("Extension is not supported")

        # download to a local folder
        remote_loaders.download(rcdb_data['backup'], filepath)

    elif uri_validator(argument):
        # argument is a url
        remote_loaders.download(argument, filepath)

    else:
        filepath = argument

    try:
        with open(filepath, 'rb') as fp:
            definition = json.load(fp)
    except:
        raise click.ClickException(f"File is not supported. Please check it at {filepath}")

    # apply a script
    try:
        import pandas as pd
        df = pd.DataFrame(definition)
        df.to_csv(output / "definition.csv", quoting=csv.QUOTE_NONNUMERIC)
        del df
    except:
        pass

    with Pool(concurrency) as pool:
        for _ in tqdm.tqdm(
                pool.imap_unordered(partial(download_and_return_id, output=output), definition),
                desc=f"Downloading images to {str(output)}",
                total=len(definition),
        ):
            pass
        pool.close()
        pool.join()


@api.group()
def stitching():
    pass


@stitching.command(name='setup')
@click.argument("argument", type=click.STRING)
@click.option('-o', "--output", type=click.Path(exists=True, dir_okay=True, file_okay=False), default=".")
@click.option('-c', '--concurrency', default=8, type=click.INT)
@pass_rebotics_context
def stitching_setup(ctx, argument, concurrency, output):
    """
    argument - id of stitching debug data from admin, or json file
    """
    output = pathlib.Path(output).absolute()
    output.mkdir(parents=True, exist_ok=True)

    filepath = output / "definition.json"

    if argument.isidigit():
        backup_id = int(argument)
        callback_data = ctx.provider.rcdb.get_core_callback_data(backup_id)
        definition = callback_data['data']

        ctx.verbose_log("Saving definition locally")
        with open(filepath, 'w') as fp:
            json.dump(fp, definition)

    elif uri_validator(argument):
        # argument is a url
        raise click.ClickException("not supported")
    else:
        ctx.verbose_log("Assuming we have a filepath")
        filepath = argument
        try:
            with open(filepath, 'rb') as fp:
                definition = json.load(fp)
        except:
            raise click.ClickException(f"File is not supported. Please check it at {filepath}")

    if ctx.verbose:
        ctx.format_result(definition['context'])

    files_to_load = []  # pair of remote_url to local path

    for call_count, call_data in enumerate(definition['calls']):
        call_folder = output / f"call_{call_count}"
        call_folder.mkdir(parents=True, exist_ok=True)

        input_data = call_data['input']

        stitching_input = {
            **input_data,
            'frame_paths': [],
        }

        del stitching_input['frame_urls']
        del stitching_input['annotated_frames']

        # add stitching result image
        files_to_load.append([
            call_data['image_url'], call_folder / "server_stitched_image.jpeg"
        ])

        mask_od_path = call_folder / "mask_od.json"
        stitching_input['mask_od'] = mask_od_path
        files_to_load.append([
            input_data['mask_od_url'], mask_od_path
        ])

        # register input files
        frames_folder = call_folder / "frames"
        frames_folder.mkdir(parents=True, exist_ok=True)

        for frame_url in input_data['frame_urls']:
            frame_path = frames_folder / get_filename_from_url(frame_url)
            stitching_input['frame_paths'].append(frame_path)
            files_to_load.append([
                frame_url, frame_path
            ])

        frames_folder = call_folder / "annotated_frames"
        frames_folder.mkdir(parents=True, exist_ok=True)

        for frame_url in input_data['annotated_frames']:
            frame_path = frames_folder / get_filename_from_url(frame_url)
            files_to_load.append([
                frame_url, frame_path
            ])

        ctx.verbose_log(f'Creating a stitching input for run #{call_count} into {call_folder}')
        with open(str(call_folder / "stitching_input.json"), 'w') as fp:
            json.dump(stitching_input, fp)

    ctx.verbose_log(f"Registered files to downloading: {files_to_load}")

    with Pool(concurrency) as pool:
        for _ in tqdm.tqdm(
            pool.starmap(remote_loaders.download, files_to_load),
            desc=f"Downloading files to {str(output)}",
            total=len(files_to_load),
        ):
            pass
        pool.close()
        pool.join()


api.add_command(shell, 'shell')
api.add_command(roles, 'roles')
api.add_command(configure, 'configure')



import json
import logging

from django.contrib.auth import get_user_model
from django.db import models
from django.utils.six import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

logger = logging.getLogger(__name__)


@python_2_unicode_compatible
class WebhookRouter(models.Model):
    user = models.ForeignKey(get_user_model(), related_name='webhooks', on_delete=models.CASCADE)
    url = models.URLField()
    auth_token = models.CharField(max_length=64, null=True)

    created = models.DateTimeField(_('Date Created'), auto_now_add=True, editable=False)
    modified = models.DateTimeField(_('Date Modified'), auto_now=True, editable=False)

    class Meta:
        verbose_name = _('Webhook router')
        verbose_name_plural = _('Webhook routers')

    def __str__(self):
        return '#{} {}-on-{}'.format(self.id, self.user, self.url)

    def emit_message(self, component, status, body):
        message = Message.objects.create(
            component=component,
            status=status,
            body=json.dumps(body),
            receiver=self,
            delivery_status=Message.DELIVERY_STATUS_PENDING
        )
        message.send()


@python_2_unicode_compatible
class Message(models.Model):
    DELIVERY_STATUS_PENDING = 'message_pending'
    DELIVERY_STATUS_DELIVERED = 'message_delivered'
    DELIVERY_STATUS_REJECTED = 'message_rejected'
    DELIVERY_STATUS_FAILED = 'message_failed'

    PROCESSING = 'processing'
    PLANOGRAM_IMPORT = 'planogram_import'
    INTERNAL = 'internal'

    receiver = models.ForeignKey(WebhookRouter, related_name='messages', on_delete=models.CASCADE)

    created = models.DateTimeField(_('Date Created'), auto_now_add=True, editable=False)
    modified = models.DateTimeField(_('Date Modified'), auto_now=True, editable=False)

    component = models.CharField(max_length=64, choices=[
        (PROCESSING, _("Processing action")),
        (PLANOGRAM_IMPORT, _("Planogram import")),
        (INTERNAL, _("Internal"))
    ])
    status = models.CharField(max_length=32)

    body = models.TextField()

    delivery_status = models.CharField(max_length=64, choices=[
        (DELIVERY_STATUS_PENDING, _("Pending")),
        (DELIVERY_STATUS_DELIVERED, _("Delivered")),
        (DELIVERY_STATUS_REJECTED, _("Rejected")),
        (DELIVERY_STATUS_FAILED, _("Failed")),
    ])

    exception_message = models.TextField()

    class Meta:
        verbose_name = _('Message')
        verbose_name_plural = _('Messages')

    def __str__(self):
        return "Message on {} to {} with {}".format(
            self.component, self.receiver_id, self.delivery_status
        )

    def send(self):
        from .tasks import send_message
        send_message.delay(message_id=self.id)



from django.contrib import admin

from rebotics_sdk.notifications.models import WebhookRouter, Message


@admin.register(WebhookRouter)
class WebhookRouterAdmin(admin.ModelAdmin):
    list_display = [
        'id', 'user', 'created', 'url'
    ]


@admin.register(Message)
class MessageAdmin(admin.ModelAdmin):
    list_display = [
        'id', 'receiver_id', 'component', 'status', 'delivery_status'
    ]



import json
import logging
import pathlib
import re
import zipfile
from collections import namedtuple, defaultdict
from datetime import datetime
from hashlib import md5
from io import BytesIO, StringIO

import more_itertools
import tqdm

import rebotics_sdk

logger = logging.getLogger(__name__)

ClassificationEntry = namedtuple('ClassificationEntry', ['label', 'feature', 'image', 'filename', 'index'])
ImageEntry = namedtuple('ImageEntry', ['filename', 'filepath', 'order'])


class VirtualClassificationEntry(object):
    def __init__(self, label, feature, image_url, index=None):
        self.label = label
        self.feature = feature
        self.image_url = image_url
        self.index = index

    def to_dict(self):
        return {
            'label': self.label,
            'feature': self.feature,
            'image_url': self.image_url,
            'index': self.index
        }

    @classmethod
    def from_dict(cls, d):
        return cls(
            label=d['label'],
            feature=d['feature'],
            image_url=d['image_url'],
            index=d['index']
        )


NUMERIC_VALUE = re.compile(r'\d+')


class ClassificationDatabaseException(Exception):
    pass


class DuplicateFeatureVectorsException(ClassificationDatabaseException):
    def __init__(self, msg, duplicates):
        super(DuplicateFeatureVectorsException, self).__init__(msg)
        self.duplicates = duplicates


def extract_numeric(filename):
    found = NUMERIC_VALUE.findall(filename)
    if len(found) == 0:
        raise ClassificationDatabaseException('Images name should contain numeric value that represents the order. '
                                              'Instead of {} name was used'.format(filename))
    return int(found[0])


class BaseClassificationDatabasePacker(object):
    version = 0
    extension = 'zip'

    def __init__(self, source=None, destination=None, progress_bar=False, check_duplicates=False):
        if source is not None and destination is not None:
            raise ValueError("You can't sent source and destination at a same time.")

        self.source = source

        self.meta_data = {
            'packed': datetime.now().strftime('%c'),
            'sdk_version': rebotics_sdk.__version__,
            'packer_version': self.version
        }
        self.zip_io = None

        if self.source is not None:
            self.read_meta()

        self.check_duplicates = check_duplicates
        self.images = []

        if hasattr(destination, 'write') or hasattr(destination, 'read'):
            self.destination = destination  # file like object
        elif destination is None:
            self.destination = BytesIO()
        elif isinstance(destination, str) or isinstance(destination, pathlib.PurePath):
            # in python 2 it doesn't accept pathlib PurePath, but only os.path or str
            self.destination = str(destination)
            if not self.destination.endswith(self.extension):
                self.destination = "{}.{}".format(self.destination, self.extension)

        self.progress_bar = progress_bar

    def read_lines(self, lines):
        return [x for x in lines.split('\n') if x]

    def zipfile(self, file, **kwargs):
        # if python version is
        params = dict(
            compression=zipfile.ZIP_DEFLATED,
            allowZip64=True,
        )
        #
        # if (sys.version_info.major, sys.version_info.minor) >= (3, 7):
        #     # using the most aggressive compression
        #     params['compresslevel'] = 9  # default is 6

        params.update(kwargs)

        for compression_option in [zipfile.ZIP_DEFLATED, zipfile.ZIP_STORED]:
            try:
                params['compression'] = compression_option
                return zipfile.ZipFile(file, **params)
            except RuntimeError:
                pass

    def __enter__(self):
        if self.zip_io is not None:
            return self

        # check for file to be not closed

        if self.source is not None:
            self.zip_io = self.zipfile(self.source, mode='r')
        elif self.destination is not None:
            self.zip_io = self.zipfile(self.destination, mode='w')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.zip_io.close()
        self.zip_io = None

    def pack(self, *args, **kwargs):
        raise NotImplementedError()

    def unpack(self):
        raise NotImplementedError()

    def read_meta(self):
        with self as packer:
            self.meta_data = json.load(packer.zip_io.open('meta.json'))


class ZipDatabasePacker(BaseClassificationDatabasePacker):
    """
    Unified file format for packing classification database into single folder
    File format is the zip based archive with following :

    labels.txt
    features.txt
    meta.json
    """
    version = 0
    extension = 'zip'

    def pack(self, labels, features, *args, **kwargs):
        labels_io = StringIO("\n".join(labels))
        features_io = StringIO("\n".join(features))
        self.meta_data['count'] = len(labels)

        with self.zipfile(self.destination, mode='w') as zip_io:
            zip_io.writestr('labels.txt', labels_io.getvalue())
            zip_io.writestr('features.txt', features_io.getvalue())
            zip_io.writestr('meta.json', json.dumps(self.meta_data))
        return self.destination

    def unpack(self):
        if self.source is None:
            raise ClassificationDatabaseException("Can not unpack with empty source file")

        with self.zipfile(self.source, mode='r') as zip_io:
            try:
                self.meta_data = json.load(zip_io.open('meta.json'))
            except KeyError:
                # meta.json is not presented, working in the compatibility mode
                pass

            labels = self.read_lines(zip_io.read('labels.txt').decode('utf-8'))
            features = self.read_lines(zip_io.read('features.txt').decode('utf-8'))

            for index, (label, feature) in enumerate(zip(labels, features)):
                yield ClassificationEntry(
                    label, feature, None, None, index
                )


class ClassificationDatabasePacker(BaseClassificationDatabasePacker):
    """
    Unified file format for packing classification database into single folder
    File format is the zip based archive with following :

    labels.txt
    features.txt
    meta.json
    images/  - folder with images, this folder can be empty
    """

    version = 1
    extension = 'rcdb'
    images_extensions = {'jpeg', 'png', 'jpg'}

    def extract_meta_data(self, labels_path, features_path, images_folder):
        known_order = set()
        for f in pathlib.Path(images_folder).iterdir():
            if f.is_file():
                extension = f.suffix
                if extension in self.images_extensions:
                    raise ClassificationDatabaseException("Extension is not supported by packer: {}".format(f.name))
                numeric_order = extract_numeric(f.name)
                if numeric_order in known_order:
                    raise ClassificationDatabaseException("Numeric order collision. {}".format(f.name))

                known_order.add(numeric_order)
                self.images.append(ImageEntry(f.name, str(f), numeric_order))

        # repack based on the filename and it's numeric value
        self.images.sort(key=lambda i: i.order)  # sort by numeric value of the filename

        with open(labels_path, 'r') as labels_io, open(features_path, 'r') as features_io:
            labels = self.read_lines(labels_io.read())
            features = self.read_lines(features_io.read())

        labels_count = len(labels)
        features_count = len(features)
        files_count = len(self.images)

        if labels_count != features_count:
            raise ClassificationDatabaseException('Inconsistent count of labels and features. {}/{}'.format(
                labels_count, features_count
            ))

        if labels_count != files_count:
            raise ClassificationDatabaseException(
                'Inconsistent count of labels and features and files. {}/{}/{}'.format(
                    labels_count, features_count, files_count
                ))

        self.meta_data.update({
            'count': features_count,
            'images': [str(img.filename) for img in self.images]
        })

        if not self.check_duplicates:
            return

        # checking for the uniqueness of the FV
        features_map = defaultdict(list)
        for i, fv in tqdm.tqdm(enumerate(features), total=features_count, disable=not self.progress_bar, leave=False):
            md_ = md5(fv.encode('utf-8'))
            features_map[md_.hexdigest()].append(i)
        duplicates = []
        for same_fv_ids in features_map.values():
            if len(same_fv_ids) > 1:
                duplicates.append([
                    ClassificationEntry(
                        labels[id_],
                        features[id_],
                        None,
                        self.images[id_].filename,
                        id_
                    ) for id_ in same_fv_ids
                ])

        if duplicates:
            raise DuplicateFeatureVectorsException(
                "Detected {} duplicate groups.".format(len(duplicates)),
                duplicates=duplicates
            )

    def pack(self, labels_path, features_path, images_folder):
        self.extract_meta_data(labels_path, features_path, images_folder)

        with self.zipfile(self.destination, mode='w') as zip_io:
            zip_io.write(labels_path, 'labels.txt')
            zip_io.write(features_path, 'features.txt')
            zip_io.writestr('meta.json', json.dumps(self.meta_data))

            if not self.progress_bar:
                for img in self.images:
                    zip_io.write(img.filepath, str(pathlib.Path('images') / img.filename))
            else:
                for img in tqdm.tqdm(self.images, leave=False):
                    zip_io.write(img.filepath, str(pathlib.Path('images') / img.filename))

        return self.destination

    def unpack(self):
        """

        :return: generator of label, feature, image_io
        :rtype:self.meta_data
        """
        if self.source is None:
            raise ClassificationDatabaseException("Can not unpack with empty source file")

        with self.zipfile(self.source, mode='r') as zip_io:
            labels = self.read_lines(zip_io.read('labels.txt').decode('utf-8'))
            features = self.read_lines(zip_io.read('features.txt').decode('utf-8'))

            for index, (label, feature, image_name) in enumerate(zip(
                labels,
                features,
                self.meta_data['images']
            )):
                image = zip_io.read('images/{}'.format(image_name))
                yield ClassificationEntry(
                    label, feature, image, image_name, index
                )


class VirtualClassificationDatabasePacker(BaseClassificationDatabasePacker):
    extension = 'vrcdb'
    version = 2
    batch_size = 10000  # smaller json files for the batches, less memory used to unmarshal it

    def __init__(self, *args, **kwargs):
        super(VirtualClassificationDatabasePacker, self).__init__(*args, **kwargs)

    def unpack(self):
        if self.source is None:
            raise ClassificationDatabaseException("Can not unpack with empty source file")

        with self as packer:
            for batch in self.meta_data['batches']:
                for entry in json.load(packer.zip_io.open(batch)):
                    yield VirtualClassificationEntry.from_dict(entry)

    def pack(self, entries):
        """

        :param list[VirtualClassificationEntry] entries: an iterable that can be passed
        :return:
        """
        with self as packer:

            for batch_number, chunk in enumerate(more_itertools.ichunked(enumerate(entries, start=1), self.batch_size)):
                data = []

                for i, entry in chunk:
                    entry.index = i
                    data.append(entry.to_dict())

                batch_filename = 'batches/batch_{}.json'.format(batch_number)
                packer.zip_io.writestr(batch_filename, json.dumps(data))

                batches = self.meta_data.get('batches', [])
                batches.append(batch_filename)
                self.meta_data['batches'] = batches

            self.meta_data['count'] = i
            packer.zip_io.writestr('meta.json', json.dumps(self.meta_data))

        return self.destination



import logging

from rest_framework import status
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from rebotics_sdk.notifications.models import WebhookRouter
from rebotics_sdk.notifications.serializers import SetWebhookSerializer
from rebotics_sdk.notifications.utils import log_action

logger = logging.getLogger(__name__)


class SetWebhookView(APIView):
    permission_classes = [
        IsAuthenticated,
    ]

    def post(self, request, *args, **kwargs):
        serializer = SetWebhookSerializer(data=request.data)

        serializer.is_valid(raise_exception=True)
        url = serializer.validated_data['url']
        user = request.user
        router, created = WebhookRouter.objects.get_or_create(user=user, url=url)
        if created:
            log_action(user, router, "Created from API")

        auth_token = serializer.validated_data.get('auth_token')
        if auth_token is not None and auth_token != router.auth_token:
            logger.debug('Overriding auth token')
            router.auth_token = auth_token
            router.save()

        logger.info('Webhook subscriber is %s for user %s with id %s', 'created' if created else 'changed', user,
                    router)
        logger.debug("Sending pong message to new webhook...")
        router.emit_message('internal', 'succeed', {
            'message': "Pong!",
        })
        return Response({
            'id': router.id,
        }, status=status.HTTP_200_OK)



import pathlib
import tempfile

import pytest

from rebotics_sdk.advanced.packers import ClassificationDatabasePacker, ZipDatabasePacker, \
    DuplicateFeatureVectorsException, VirtualClassificationDatabasePacker, VirtualClassificationEntry


@pytest.fixture(scope="module")
def script_cwd(request):
    return request.fspath.join("..")


def test_classification_packing(script_cwd):
    db_folder = pathlib.Path(script_cwd.join("db"))
    with tempfile.TemporaryDirectory() as dirname:
        destination_filename = pathlib.Path(dirname, 'test')
        packer = ClassificationDatabasePacker(destination=destination_filename)
        features = db_folder / 'features.txt'
        labels = db_folder / 'labels.txt'
        images_folder = db_folder / 'custom_folder/'

        res = packer.pack(labels, features, images_folder)

        assert 'test.rcdb' in res
        assert len(packer.images) == 2

        packer = ClassificationDatabasePacker(source=res)
        entries = list(packer.unpack())
        assert len(entries) == 2
        entry = entries[0]
        assert entry.label == '123123123'
        assert entry.feature == '123123123123123'
        internal_filename = entry.filename
        assert internal_filename == 'image_1.png'

        # testing if it can be dumped to the FS
        og_file = images_folder / internal_filename
        tmp_file = db_folder / internal_filename

        with open(tmp_file, 'wb') as fout:
            fout.write(entry.image)

        assert og_file.stat().st_size == tmp_file.stat().st_size
        tmp_file.unlink()


def test_classification_packing_check_duplicates(script_cwd):
    db_folder = pathlib.Path(script_cwd.join("db"))
    packer = ClassificationDatabasePacker(destination='test', check_duplicates=True)
    features = db_folder / 'features.txt'
    labels = db_folder / 'labels.txt'
    images_folder = db_folder / 'custom_folder/'

    with pytest.raises(DuplicateFeatureVectorsException) as excinfo:
        packer.pack(labels, features, images_folder)
    assert "duplicate" in str(excinfo.value)


def test_zip_packing():
    packer = ZipDatabasePacker()
    packed = packer.pack(
        labels=[
            '123123123'
        ],
        features=[
            '123123123123123'
        ]
    )
    assert packer.meta_data['count'] == 1

    unpacker = ZipDatabasePacker(source=packed)
    for entry in unpacker.unpack():
        assert entry.label == '123123123'
        assert entry.feature == '123123123123123'

    assert unpacker.meta_data['count'] == 1


def test_virtual_packing():
    packer = VirtualClassificationDatabasePacker()
    packed = packer.pack([
        VirtualClassificationEntry(
            label='{}'.format(i),
            feature='{}'.format(i) * 5,
            image_url='https://example.com/image_{}.jpeg'.format(i)
        ) for i in range(100)
    ])
    assert packer.meta_data['count'] == 100
    assert len(packer.meta_data['batches']) == 1

    unpacker = VirtualClassificationDatabasePacker(source=packed)
    assert unpacker.meta_data['count'] == 100
    for i, entry in enumerate(unpacker.unpack()):
        if i == 0:
            first_entry = entry
            assert first_entry.label == '0'
            assert first_entry.feature == '00000'
            assert first_entry.index == 1
            assert first_entry.image_url == 'https://example.com/image_0.jpeg'
            break
